
#include <sys/param.h>
#include <sys/stat.h>

#include <ctype.h>
#include <err.h>
#include <paths.h>
#include <pwd.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <util.h>

#include "xpw.h"

struct gecos {
	char	*full_name;
	char	*office_location;
	char	*office_phone;
	char	*home_phone;
};

static char	 tempfile[] = __CONCAT(_PATH_VARTMP,"pw.XXXXXXXX");

static void
usage(void)
{
	extern char *__progname;
	
	fprintf(stderr, "Usage: %s\n", __progname);
	exit(1);
}

char *
ok_shell(char *name)
{
	char *p, *sh;
	
	setusershell();
	while ((sh = getusershell()) != NULL) {
		if (!strcmp(name, sh))
			return (name);
		/* allow just shell name, but use "real" path */
		if ((p = strrchr(sh, '/')) && strcmp(name, p + 1) == 0)
			return (sh);
	}
	return (NULL);
}

static void
cleanup_tempfile(void)
{
	unlink(tempfile);
}

static void
kbintr(int sig)
{
	char msg[] = "\nUser information unchanged.\n";

	cleanup_tempfile();
        write(STDERR_FILENO, msg, sizeof(msg) - 1);
        _exit(1);
}

int
main(int argc, char *argv[])
{
	struct passwd *pw;
	struct stat begin, end;
	struct gecos gc;
	struct xpw req, rep;
	FILE *fp;
	char *field, *value, shell[MAXPATHLEN], buf[BUFSIZ];
	int i, fd;
	
	if (argc != 1)
		usage();

	if ((pw = getpwuid(getuid())) == NULL)
		err(1, "unknown user %u", getuid());

	/* Create our temporary update file. */
	if ((fd = mkstemp(tempfile)) < 0)
		err(1, "couldn't create temporary file");
	
	if ((fp = fdopen(fd, "w")) == NULL)
		err(1, "fdopen");
	
	atexit(cleanup_tempfile);
	signal(SIGINT, kbintr);
	signal(SIGTERM, kbintr);
	
	/* Write current user information to update file. */
	fprintf(fp, "# Changing user database information for %s.\n",
	    pw->pw_name);

	if (ok_shell(pw->pw_shell)) {
		fprintf(fp, "Shell: %s\n",
		    *pw->pw_shell ? pw->pw_shell : _PATH_BSHELL);
	}
	field = pw->pw_gecos;
	value = strsep(&field, ",");
	fprintf(fp, "Full Name: %s\n", value ? value : "");
	value = strsep(&field, ",");
	fprintf(fp, "Office Location: %s\n", value ? value : "");
	value = strsep(&field, ",");
	fprintf(fp, "Office Phone: %s\n", value ? value : "");
	value = strsep(&field, ",");
	fprintf(fp, "Home Phone: %s\n", value ? value : "");

	fclose(fp);

	/* Edit the update file. */
	if (lstat(tempfile, &begin) < 0 || S_ISLNK(begin.st_mode))
		err(1, "error editing %s", tempfile);
	pw_edit(0, tempfile);
	if (lstat(tempfile, &end) < 0 || S_ISLNK(end.st_mode))
		err(1, "error editing %s", tempfile);
	
	if (begin.st_mtime == end.st_mtime && begin.st_size == end.st_size)
		errx(0, "no changes made");
	
	/* Parse update. */
	if ((fp = fopen(tempfile, "r")) == NULL)
		err(1, "couldn't open %s", tempfile);

	for (i = 0; fgets(buf, sizeof(buf), fp) != NULL; i++) {
		if (isspace((int)buf[0]) || buf[0] == '#')
			continue;

		value = buf;
		if ((field = strsep(&value, ":")) == NULL || value == NULL)
			errx(1, "corrupted input on line %d", i);

		while (isspace((int)value[0]))
			value++;
		strtok(value, "\r\n");

		if (strpbrk(value, ":,") != NULL)
			errx(1, "illegal character in the \"%s\" field",
			    field);
		
		if (strcasecmp(field, "Shell") == 0) {
			if (!ok_shell(value))
				errx(1, "%s: non-standard shell", value);
			strlcpy(shell, value, sizeof(shell));
		} else if (strcasecmp(field, "Full Name") == 0) {
			gc.full_name = strdup(value);
		} else if (strcasecmp(field, "Office Location") == 0) {
			gc.office_location = strdup(value);
		} else if (strcasecmp(field, "Office Phone") == 0) {
			gc.office_phone = strdup(value);
		} else if (strcasecmp(field, "Home Phone") == 0) {
			gc.home_phone = strdup(value);
		} else
			errx(1, "unrecognized field \"%s\" on line %d",
			    field, i);
	}
	fclose(fp);
	
	if (snprintf(buf, sizeof(buf), "%s,%s,%s,%s", gc.full_name,
	    gc.office_location, gc.office_phone, gc.home_phone) >=
	    sizeof(buf) - 1)
		errx(1, "entries too long");
	
	/* Request update. */
	req.xpw_type = XPW_REQ_CHPASS;
	req.xpw_shell = shell;
	req.xpw_gecos = buf;
	
	if (xpw_call(&req, &rep) < 0)
		err(1, "couldn't contact password daemon");
	
	if (rep.xpw_type == XPW_REP_ERR)
		errx(1, "%s", rep.xpw_reply);
	
	printf("User information updated.\n");
	
	exit(0);
}
