
#include <sys/types.h>

#include <err.h>
#include <errno.h>
#include <login_cap.h>
#include <pwd.h>
#include <readpassphrase.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "xpw.h"

extern int	 pwd_check(struct passwd *pwd, login_cap_t *lc, char *pass);
extern int	 pwd_gettries(struct passwd *pwd, login_cap_t *lc);

static void
usage(void)
{
	extern char *__progname;
	
	fprintf(stderr, "Usage: %s\n", __progname);
	exit(1);
}

static void
kbintr(int sig)
{
	char msg[] = "\nPassword unchanged.\n";
	
        write(STDERR_FILENO, msg, sizeof(msg) - 1);
        _exit(1);
}

int
main(int argc, char *argv[])
{
	struct passwd *pw;
	struct xpw req, rep;
	login_cap_t *lc;
	char oldpass[_PASSWORD_LEN], newpass[_PASSWORD_LEN];
	int i, pwd_tries, status;
	
	if (argc != 1)
		usage();

	if ((pw = getpwuid(getuid())) == NULL)
		err(1, "unknown user %u", getuid());
	
	if ((lc = login_getclass(pw->pw_class)) == NULL)
		err(1, "unable to get login class for %s.", pw->pw_name);
	
	printf("Changing local password for %s.\n", pw->pw_name);

	signal(SIGHUP, kbintr);
	signal(SIGINT, kbintr);
	signal(SIGTERM, kbintr);
	
	/* Get old password. */
	req.xpw_type = XPW_REQ_PASSWD;
	if ((req.xpw_oldpass = readpassphrase("Old password:", oldpass,
	    sizeof(oldpass), 0)) == NULL) {
		printf("Password unchanged.\n");
		exit(1);
	}
	/* Get new password. */
	pwd_tries = pwd_gettries(pw, lc);
	
	for (i = 0; i < pwd_tries || pwd_tries == 0; i++) {
		if ((req.xpw_newpass = readpassphrase("New password:",
		    newpass, sizeof(newpass), 0)) == NULL) {
			memset(oldpass, 0, sizeof(oldpass));
			printf("Password unchanged.\n");
			exit(1);
		}
		if (strcmp(req.xpw_newpass, "s/key") == 0) {
			printf("That password collides with a system feature. "
			    "Choose another.\n");
		} else if (pwd_check(pw, lc, req.xpw_newpass)) {
			if (strcmp(req.xpw_newpass,
			    getpass("Retype new password:")) == 0)
				break;
			
			printf("Mismatch; try again, ^C to quit.\n");
		}
	}
	/* Request password change. */
	if ((status = xpw_call(&req, &rep)) == 0) {
		if (rep.xpw_type == XPW_REP_OK) {
			printf("Password changed.\n");
		} else if (rep.xpw_type == XPW_REP_ERR) {
			printf("%s\n", rep.xpw_reply);
			status = -1;
		} else {
			printf("Unexpected reply type: %d\n", req.xpw_type);
			status = -1;
		}
		xpw_xdr_free(&rep);
	} else
		printf("Password daemon unreachable: %s\n", strerror(errno));
	
	memset(oldpass, 0, sizeof(oldpass));
	memset(newpass, 0, sizeof(newpass));
	
	exit(status);
}
