
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>

#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "xpw.h"

size_t
xpw_xdr_encode(struct xpw *xpw, void *buf, size_t len)
{
	XDR xdrs;
	
	xdrmem_create(&xdrs, buf, len, XDR_ENCODE);
	len = xdr_xpw(&xdrs, xpw) ? xdr_getpos(&xdrs) : -1;
	xdr_destroy(&xdrs);
	
	return (len);
}

size_t
xpw_xdr_decode(void *buf, size_t len, struct xpw *xpw)
{
	XDR xdrs;

	memset(xpw, 0, sizeof(*xpw));
	xdrmem_create(&xdrs, buf, len, XDR_DECODE);
	len = xdr_xpw(&xdrs, xpw) ? xdr_getpos(&xdrs) : -1;
	xdr_destroy(&xdrs);
	
	return (len);
}

int
xpw_xdr_free(struct xpw *xpw)
{
	if (xpw != NULL) {
		if (xpw->xpw_oldpass != NULL)
			memset(xpw->xpw_oldpass, 0,
			    strlen(xpw->xpw_oldpass));
		if (xpw->xpw_newpass != NULL)
			memset(xpw->xpw_newpass, 0,
			    strlen(xpw->xpw_newpass));
		xdr_free(xdr_xpw, (char *)xpw);
		return (0);
	}
	return (-1);
}

int
xpw_call(struct xpw *req, struct xpw *rep)
{
	struct sockaddr_un sun;
	char buf[4096];
	int fd, len, ret = -1;

	memset(&sun, 0, sizeof(sun));
	sun.sun_len = sizeof(sun);
	sun.sun_family = AF_UNIX;
	strlcpy(sun.sun_path, _PATH_XPW_SOCK, sizeof(sun.sun_path));
	
	if ((fd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0)
		return (ret);
	
	if (connect(fd, (struct sockaddr *)&sun, sizeof(sun)) == 0 &&
	    (len = xpw_xdr_encode(req, buf, sizeof(buf))) > 0) {
		if (write(fd, buf, len) == len) {
			memset(buf, 0, len);
			
			if ((len = read(fd, buf, sizeof(buf))) > 0) {
				if ((len = xpw_xdr_decode(buf, len, rep)) > 0)
					ret = 0;
				memset(buf, 0, len);
			}
		}
	}
	close(fd);
	return (ret);
}
