
#include <fcntl.h>
#include <login_cap.h>
#include <pwd.h>
#include <string.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>
#include <util.h>

#include "xpw.h"

extern int	pwd_gensalt(char *, int, struct passwd *, login_cap_t *, char);

static int
update_passwd(struct passwd *pw, int pwflags)
{
	int i, pfd, tfd;
	
	/* Get a lock on the passwd file and open it. */
	pw_init();
	
	for (i = 1; (tfd = pw_lock(0)) == -1; i++) {
		usleep(250000);
	}
	pfd = open(_PATH_MASTERPASSWD, O_RDONLY, 0);
	
	if (pfd < 0 || fcntl(pfd, F_SETFD, 1) == -1) {
		syslog(LOG_ERR, "error opening %s: %m", _PATH_MASTERPASSWD);
		return (-1);
	}
	/* Update master.passwd file and rebuild spwd.db. */
	pw_copy(pfd, tfd, pw);
	
	if (pw_mkdb(pw->pw_name, pwflags) < 0) {
		syslog(LOG_ERR, "pw_mkdb: %m");
		return (-1);
	}
	return (0);
}

int
local_passwd(struct xpw *xpw, struct passwd *pw)
{
	char salt[_PASSWORD_LEN];
	login_cap_t *lc;
	time_t period;
	int pwflags = _PASSWORD_OMITV7;

	if (strcmp(crypt(xpw->xpw_oldpass, pw->pw_passwd),
	    pw->pw_passwd) != 0) {
		syslog(LOG_ERR, "incorrect old password for %s", pw->pw_name);
		return (-1);
	}
	if ((lc = login_getclass(pw->pw_class)) == NULL ||
	    !pwd_gensalt(salt, sizeof(salt), pw, lc, 'l'))
		return (-1);
	
	/* Get the new password. */
	pw->pw_passwd = crypt(xpw->xpw_newpass, salt);
	
	/* Reset password change time based on login.conf. */
	if ((period = login_getcaptime(lc, "passwordtime", 0, 0)) > 0) {
		pw->pw_change = time(NULL) + period;
	} else {
		/*
		 * If the pw change time is the same we only
		 * need to update the spwd.db file.
		 */
		if (pw->pw_change != 0)
			pw->pw_change = 0;
		else
			pwflags = _PASSWORD_SECUREONLY;
	}
	return (update_passwd(pw, pwflags));
}

static char *
ok_shell(char *name)
{
	char *p, *sh;

	setusershell();
	while ((sh = getusershell()) != NULL) {
		if (!strcmp(name, sh))
			return (name);
		/* allow just shell name, but use "real" path */
		if ((p = strrchr(sh, '/')) && strcmp(name, p + 1) == 0)
			return (sh);
	}
	return (NULL);
}

int
local_chpass(struct xpw *xpw, struct passwd *pw)
{
	char *p;
	int i;
	
	if (xpw->xpw_gecos == NULL || xpw->xpw_shell == NULL)
		return (-1);

	if (strpbrk(xpw->xpw_gecos, ":") != NULL ||
	    strpbrk(xpw->xpw_shell, ":,") != NULL) {
		syslog(LOG_ERR, "illegal characters in shell/GECOS "
		    "update from %s", pw->pw_name);
		return (-1);
	}
	for (i = 0, p = xpw->xpw_gecos; (p = strchr(p, ',')) != NULL; i++, p++)
		;

	if (i != 3) {
		syslog(LOG_ERR, "incorrect number (%d) of GECOS fields in "
		    "update from %s", i, pw->pw_name);
		return (-1);
	}
	if (!ok_shell(xpw->xpw_shell)) {
		syslog(LOG_ERR, "invalid shell \"%s\" in update from %s",
		    xpw->xpw_shell, pw->pw_name);
		return (-1);
	}
	pw->pw_gecos = xpw->xpw_gecos;
	pw->pw_shell = xpw->xpw_shell;
	
	return (update_passwd(pw, 0));
}
